source("code/nlp-api.R")

shhh(library(testthat))

interview_list = c("CRYSTAL", "DR. SAFAEE")
patient_dialog_id = "PATIENT"

extra_patient_respondent = c("UNIDENTIFIED WOMAN:", "UNIDENTIFIED MAN:")

sentiment_ranges = read_csv("dictionaries/dictionary_ranges.csv")
patient_quality_of_life = read_csv("data/patient_qol.csv")
question_id_key = read_csv("data/question_map.csv")

control_interviews_df = read_csv("data/control_sentiment_at_role_level.csv")
control_quality_of_life = read_csv("data/control_patient_qol.csv")
include_control_interviews = TRUE

if (include_control_interviews)
{
  patient_quality_of_life = patient_quality_of_life %>% bind_rows(control_quality_of_life)
}
get_all_files_in_data_dir = function(log_dir, log_file_regex)
{
  return(list.files(log_dir, full.names = T, pattern="*[0-9]+.pdf"))
}

range01 <- function(score, min_score, max_score){
  to_return = (score-min_score)/(max_score-min_score)
  return (to_return)
}


clean_up_text = function(text_df)
{
  text_df = text_df %>% mutate(text = text %>% str_remove_all("[0-9]+"))
  text_df = text_df %>% mutate(text = text %>% str_squish())
  text_df = text_df %>% mutate(text = text %>% toupper())
  return(text_df)
}

compute_normed_avg_sentiment = function(mins, maxes, scores)
{
  scores = unlist(scores, use.names = FALSE)
  normed_scores = c()
  
  for (i in 1:length(scores))
  {
    if (scores[i] %>% is.na() == FALSE & scores[i] %>% is.nan() == FALSE)
    {
      normed_score = range01(scores[i], mins[i], maxes[i])
      normed_scores = c(normed_scores, normed_score)
    }
  }
  if (normed_scores %>% length() > 0)
  {
    return (median(normed_scores))
  }
  else
  {
    return (NA)
  }
}

scrub_df_of_extra_patient_respondent = function(text_df)
{
  for (i in 1:length(extra_patient_respondent))
  {
    cur_extra_patient_respondent = extra_patient_respondent[i]
    text_df = text_df %>% mutate(text = text %>% str_remove_all(cur_extra_patient_respondent))
    text_df = text_df %>% mutate(text = text %>% str_squish())
  }
  return (text_df)
}

identify_interviewer = function(interview_text)
{
  interviewer_counts = rep(0, length(interview_list))
  for (i in 1:length(interview_list))
  {
    interviewer = interview_list[i]
    interviewer_counts[i] = str_count(interview_text, interviewer)
  }
  max_index = interviewer_counts %>% which.max()
  return (interview_list[max_index])
}

get_patient_id = function(document_id)
{
  patient_id = document_id %>% str_extract_all("[0-9]+")
  return (patient_id)
}

add_aggregate_get_patient_quality_of_life = function(df)
{
  dup_df = df %>% add_column(AGG = NA)
  dup_df = dup_df %>% mutate(ODI = 1-ODI)
  dup_df = dup_df %>% mutate(Frailty = 1-Frailty)
  for (i in 1:nrow(dup_df))
  {
    cur_ODI = dup_df$ODI[i]
    cur_Frailty = dup_df$Frailty[i]
    cur_SF36v2 = dup_df$SF36v2[i]
    cur_SRS22 = dup_df$SRS22[i]
    vector_of_qol_measures = c(cur_ODI, cur_Frailty, cur_SRS22, cur_SRS22)
    median_measure = median(vector_of_qol_measures, na.rm = TRUE)
    dup_df$AGG[i] = median_measure
  }
  df$AGG = dup_df$AGG
  return(df)
}

get_patient_quality_of_life_classification = function(the_patient_id, measure = "ODI_Classification")
{
  patient_id_to_look_up = as.double(as.character(the_patient_id))
  patient_row = patient_quality_of_life %>% filter(FTID == patient_id_to_look_up)
  patient_row = patient_row %>% select(all_of(measure))
  to_return = patient_row[1,1]
  return(to_return)
}

get_patient_quality_of_life = function(the_patient_id, measure = "ODI")
{
  patient_id_to_look_up = as.double(as.character(the_patient_id))
  patient_row = patient_quality_of_life %>% filter(FTID == patient_id_to_look_up)
  patient_row = patient_row %>% select(all_of(measure))
  to_return = patient_row[1,1]
  return(to_return)
}

delineate_text = function(text_df)
{
  text_df$interviewer_text = NA
  text_df$patient_text = NA
  
  for (i in 1:nrow(text_df))
  {
    success = FALSE
    text_row = text_df$text[i]
    interviewer = text_df$interviewer[i]
    try({
      text_df$interviewer_text[i] = delineate_interviewer_text(text_row, interviewer)
      text_df$patient_text[i] = delineate_patient_text(text_row, interviewer)
      success = TRUE
    })
    if (success == FALSE)
    {
      print("________________________________________________")
      print("PROBLEM WITH THE FOLLOWING")
      print(paste("PATIENT ID", patient_uid))
      print(paste("INTERVIEWER", text_interviewer))
      print("________________________________________________")
    }
  }
  return(text_df)
}

handle_misaligned_patient_locs = function(text, patient_token_locations, interview_token_locations)
{
  patient_token_locations = handle_too_long_patient_locs(text, patient_token_locations, interview_token_locations)
  patient_token_locations = handle_too_short_patient_locs(text, patient_token_locations, interview_token_locations)
  return (patient_token_locations)
  
}

transform_to_volley_df = function(text_df)
{
  volley_df = tibble(document=character(), patient_id = character(), volley_id=integer(), interviewer_volley_text=character(), patient_volley_text=character())
  
  for (i in 1:nrow(text_df))
  {
    text_row = text_df$text[i]
    interviewer = text_df$interviewer[i]
    document_id = text_df$document[i]
    buf_patient_id = text_df$patient[i]
    interviewer_volleys = delineate_interviewer_text(text_row, interviewer, get_volleys=TRUE)
    patient_volleys = delineate_patient_text(text_row, interviewer, get_volleys=TRUE)
    
    for (j in 1:length(interviewer_volleys))
    {
      cur_interviewer_volley_text = interviewer_volleys[j]
      cur_patient_volley_text = patient_volleys[j]
      volley_df = volley_df %>% add_row(document=document_id,
                                        patient_id=buf_patient_id,
                                        volley_id=j, 
                                        interviewer_volley_text = cur_interviewer_volley_text, 
                                        patient_volley_text=cur_patient_volley_text)
    }
  }
  return (volley_df)
}

analyze_sentiment_from_sa_package = function(text_df, text, colname_prefix, include_only=c("qdap", "lm", "gi", "he"))
{
  include_all=c("qdap", "lm", "gi", "he")
  include_only = paste0(colname_prefix, "_", include_only)
  reference_df = text %>% analyzeSentiment()
  colnames_to_use = paste0(colname_prefix, "_", include_all)
  to_return = cbind.data.frame(reference_df$SentimentQDAP, reference_df$SentimentLM, reference_df$SentimentGI, reference_df$SentimentHE)
  colnames(to_return) = colnames_to_use
  to_return = to_return %>% select(contains(include_only))
  to_return = text_df %>% cbind.data.frame(to_return)
  return(to_return)
}

score_qdap_sentiment = function(text)
{
  sentiment_buffer_obj = text %>% analyzeSentiment()
  return (sentiment_buffer_obj$SentimentQDAP)
}

score_lm_sentiment = function(text)
{
  sentiment_buffer_obj = text %>% analyzeSentiment()
  return (sentiment_buffer_obj$SentimentLM)
}

score_he_sentiment = function(text)
{
  sentiment_buffer_obj = text %>% analyzeSentiment()
  return (sentiment_buffer_obj$SentimentHE)
}

score_gi_sentiment = function(text)
{
  sentiment_buffer_obj = text %>% analyzeSentiment()
  return (sentiment_buffer_obj$SentimentGI)
}

handle_too_short_patient_locs = function(text, patient_token_locations, interview_token_locations)
{
  if (patient_token_locations %>% length() < interview_token_locations %>% length())
  {
    to_return = patient_token_locations
    to_return = c(to_return, nchar(text)+1)
    return (to_return)
  }
  return(patient_token_locations)
}

get_max_sentiment_scores = function(sentiment_score_names)
{
  max_sentiment_scores = c()
  for (i in 1:length(sentiment_score_names))
  {
    sentiment_range_row = sentiment_ranges %>% filter(sentiment_analysis_method == sentiment_score_names[i])
    max_score_for_method = sentiment_range_row$max_score[1]
    max_sentiment_scores = c(max_sentiment_scores, max_score_for_method)
  }
  return (max_sentiment_scores)
}

get_min_sentiment_scores = function(sentiment_score_names)
{
  min_sentiment_scores = c()
  for (i in 1:length(sentiment_score_names))
  {
    sentiment_range_row = sentiment_ranges %>% filter(sentiment_analysis_method == sentiment_score_names[i])
    min_score_for_method = sentiment_range_row$min_score[1]
    min_sentiment_scores = c(min_sentiment_scores, min_score_for_method)
  }
  return (min_sentiment_scores)
}

norm_sentiment_scores_for_interviewer = function(df, df_type = "fulltext")
{
  sentiment_df = df %>% select(contains("sentiment"))
  sentiment_df = sentiment_df %>% select(contains("interviewer"))
  sentiment_score_names = colnames(sentiment_df)
  sentiment_score_names = sentiment_score_names %>% str_remove_all("interviewer_")
  sentiment_score_names = sentiment_score_names %>% str_remove_all("sentiment_")
  
  
  
  if (df_type == "fulltext")
  {
    df$interviewer_sentiment_agg=NA
    max_sentiment_scores = get_max_sentiment_scores(sentiment_score_names)
    min_sentiment_scores = get_min_sentiment_scores(sentiment_score_names)
    for (i in 1:nrow(df))
    {
      vector_of_scores = sentiment_df[i,]
      df$interviewer_sentiment_agg[i] = compute_normed_avg_sentiment(min_sentiment_scores, max_sentiment_scores, vector_of_scores)
    }
  }
  else if (df_type == "volley")
  {
    df$interviewer_volley_sentiment_agg=NA
    sentiment_score_names = sentiment_score_names %>% str_remove_all("volley_")
    max_sentiment_scores = get_max_sentiment_scores(sentiment_score_names)
    min_sentiment_scores = get_min_sentiment_scores(sentiment_score_names)
    for (i in 1:nrow(df))
    {
      vector_of_scores = sentiment_df[i,]
      df$interviewer_volley_sentiment_agg[i] = compute_normed_avg_sentiment(min_sentiment_scores, max_sentiment_scores, vector_of_scores)
    }
  }
  else if (df_type == "question")
  {
    df$interviewer_sentiment_agg=NA
    sentiment_score_names = sentiment_score_names %>% str_remove_all("question_")
    max_sentiment_scores = get_max_sentiment_scores(sentiment_score_names)
    min_sentiment_scores = get_min_sentiment_scores(sentiment_score_names)
    for (i in 1:nrow(df))
    {
      vector_of_scores = sentiment_df[i,]
      df$interviewer_sentiment_agg[i] = compute_normed_avg_sentiment(min_sentiment_scores, max_sentiment_scores, vector_of_scores)
    }
  }
  return(df)
}

norm_sentiment_scores_for_patient = function(df, df_type = "fulltext")
{
  sentiment_df = df %>% select(contains("sentiment"))
  sentiment_df = sentiment_df %>% select(contains("patient"))
  sentiment_score_names = colnames(sentiment_df)
  sentiment_score_names = sentiment_score_names %>% str_remove_all("patient_")
  sentiment_score_names = sentiment_score_names %>% str_remove_all("sentiment_")
  
  
  
  if (df_type == "fulltext")
  {
    df$patient_sentiment_agg=NA
    max_sentiment_scores = get_max_sentiment_scores(sentiment_score_names)
    min_sentiment_scores = get_min_sentiment_scores(sentiment_score_names)
    for (i in 1:nrow(df))
    {
      vector_of_scores = sentiment_df[i,]
      df$patient_sentiment_agg[i] = compute_normed_avg_sentiment(min_sentiment_scores, max_sentiment_scores, vector_of_scores)
    }
  }
  else if (df_type == "volley")
  {
    df$patient_volley_sentiment_agg=NA
    sentiment_score_names = sentiment_score_names %>% str_remove_all("volley_")
    max_sentiment_scores = get_max_sentiment_scores(sentiment_score_names)
    min_sentiment_scores = get_min_sentiment_scores(sentiment_score_names)
    for (i in 1:nrow(df))
    {
      vector_of_scores = sentiment_df[i,]
      df$patient_volley_sentiment_agg[i] = compute_normed_avg_sentiment(min_sentiment_scores, max_sentiment_scores, vector_of_scores)
    }
  }
  else if (df_type == "question")
  {
    df$patient_answer_sentiment_agg=NA
    sentiment_score_names = sentiment_score_names %>% str_remove_all("answer_")
    max_sentiment_scores = get_max_sentiment_scores(sentiment_score_names)
    min_sentiment_scores = get_min_sentiment_scores(sentiment_score_names)
    for (i in 1:nrow(df))
    {
      vector_of_scores = sentiment_df[i,]
      df$patient_answer_sentiment_agg[i] = compute_normed_avg_sentiment(min_sentiment_scores, max_sentiment_scores, vector_of_scores)
    }
  }
  return(df)
}

handle_too_long_patient_locs = function(text, patient_token_locations, interview_token_locations)
{
  if (patient_token_locations %>% length() > interview_token_locations %>% length())
  {
    to_return = c()
    for (i in 1:length(interview_token_locations))
    {
      cur_interview_token_loc = interview_token_locations[i]
      for (j in 1:length(patient_token_locations))
      {
        cur_patient_token_loc = patient_token_locations[j]
        if (cur_patient_token_loc > cur_interview_token_loc)
        {
          to_return = c(to_return, cur_patient_token_loc)
          break
        }
      }
    }
    return (to_return)
  }
  return (patient_token_locations)
}

assign_interviewer_question_id = function(text)
{
  if (text %>% is.na())
  {
    return(NA)
  }
  for (i in 1:nrow(question_id_key))
  {
    cur_distinctive_text = question_id_key$distinctive_phrase[i]
    cur_question_id = question_id_key$question_id[i]
    
    if (text %>% str_detect(cur_distinctive_text))
    {
      return(cur_question_id)
    }
  }
  return(NA)
}

delineate_interviewer_text = function(text, interviewer, get_volleys=FALSE)
{
  length_of_text = nchar(text)
  interviewer_token = paste0(interviewer, ":")
  patient_token = paste0(patient_dialog_id, ":")
  interview_token_locations = str_locate_all(text, interviewer_token)
  patient_token_locations = str_locate_all(text, patient_token)
  
  interview_token_locations = interview_token_locations[[1]]
  interview_token_locations = interview_token_locations[,2]
  
  patient_token_locations = patient_token_locations[[1]]
  patient_token_locations = patient_token_locations[,1]
  
  patient_token_locations = handle_misaligned_patient_locs(text, patient_token_locations, interview_token_locations)
  
  expect_equal(interview_token_locations %>% length(), patient_token_locations %>% length())
  
  interviewer_text = ""
  interview_text_vector = c()
  for (i in 1:length(interview_token_locations))
  {
    start_location = interview_token_locations[i]+1
    stop_location = patient_token_locations[i]-1
    
    interview_text_vector = c(interview_text_vector, text %>% substring(start_location, stop_location) %>% str_squish())
    interviewer_text = paste(interviewer_text, text %>% substring(start_location, stop_location))
    interviewer_text = interviewer_text %>% str_squish()
  }
  if (get_volleys)
  {
    return (interview_text_vector)
  }
  else{
    return (interviewer_text)
  }
  
}

delineate_patient_text = function(text, interviewer, get_volleys=FALSE)
{
  length_of_text = nchar(text)
  interviewer_token = paste0(interviewer, ":")
  patient_token = paste0(patient_dialog_id, ":")
  interview_token_locations = str_locate_all(text, interviewer_token)
  patient_token_locations = str_locate_all(text, patient_token)
  
  interview_token_locations = interview_token_locations[[1]]
  interview_token_locations = interview_token_locations[,1]
  
  patient_token_locations = patient_token_locations[[1]]
  patient_token_locations = patient_token_locations[,2]
  
  
  if (patient_token_locations %>% length() > interview_token_locations %>% length())
  {
    new_patient_token_locs = c()
    for (i in 1:length(interview_token_locations))
    {
      cur_interview_token_loc = interview_token_locations[i]
      for (j in 1:length(patient_token_locations))
      {
        cur_patient_token_loc = patient_token_locations[j]
        if (cur_patient_token_loc > cur_interview_token_loc)
        {
          new_patient_token_locs = c(new_patient_token_locs, cur_patient_token_loc)
          break
        }
      }
    }
    patient_token_locations = new_patient_token_locs
  }
  
  patient_token_locations = handle_misaligned_patient_locs(text, patient_token_locations, interview_token_locations)
  
  expect_equal(interview_token_locations %>% length(), patient_token_locations %>% length())
  
  patient_text = ""
  patient_text_vector = c()
  for (i in 1:length(patient_token_locations))
  {
    start_location = patient_token_locations[i]+1
    stop_location = interview_token_locations[i+1]-1
    
    patient_text_vector = c(patient_text_vector, text %>% substring(start_location, stop_location) %>% str_squish())
    patient_text = paste(patient_text, text %>% substring(start_location, stop_location))
    patient_text = patient_text %>% str_squish()
  }
  if (get_volleys)
  {
    return (patient_text_vector)
  }
  else
  {
    return (patient_text )
  }
  
}

data_pdf_files = get_all_files_in_data_dir("data")
text_df = from.a.list.of.files.to.file.text.df(data_pdf_files)
text_df = text_df %>% group_by(document) %>% summarise(text=paste(text,collapse=' '))
if (include_control_interviews)
{
  text_df = text_df %>% bind_rows(control_interviews_df)
}
text_df = text_df %>% clean_up_text()
text_df = text_df %>% scrub_df_of_extra_patient_respondent()
text_df$interviewer = unlist(lapply(text_df$text, identify_interviewer))
text_df$patient = unlist(lapply(text_df$document, get_patient_id))

patient_quality_of_life = patient_quality_of_life %>% add_aggregate_get_patient_quality_of_life()

text_df$patient_quality_of_life_score_odi = unlist(lapply(text_df$patient, get_patient_quality_of_life))
text_df$patient_quality_of_life_classification_odi = unlist(lapply(text_df$patient, get_patient_quality_of_life_classification))
text_df$patient_quality_of_life_score_frailty = unlist(lapply(text_df$patient, get_patient_quality_of_life, measure="Frailty"))
text_df$patient_quality_of_life_classification_frailty = unlist(lapply(text_df$patient, get_patient_quality_of_life_classification, measure="Frailty_Classification"))
text_df$patient_quality_of_life_score_sf36v2 = unlist(lapply(text_df$patient, get_patient_quality_of_life, measure="SF36v2"))
text_df$patient_quality_of_life_score_srs22 = unlist(lapply(text_df$patient, get_patient_quality_of_life, measure="SRS22"))
text_df$patient_quality_of_life_score_agg = unlist(lapply(text_df$patient, get_patient_quality_of_life, measure="AGG"))
text_df = text_df %>% delineate_text()

text_volley_df = text_df %>% transform_to_volley_df()
text_volley_df$question_id = unlist(lapply(text_volley_df$interviewer_volley_text, assign_interviewer_question_id))
text_volley_df = text_volley_df %>% fill(question_id)

text_question_df = text_volley_df
text_question_df = text_question_df %>% group_by(patient_id, question_id) %>% mutate(patient_answer_text = paste0(patient_volley_text, collapse = " "))
text_question_df = text_question_df %>% ungroup()
text_question_df = text_question_df %>% group_by(patient_id, question_id) %>% mutate(interviewer_question_text = paste0(interviewer_volley_text, collapse = " "))
text_question_df = text_question_df %>% ungroup()
text_question_df = text_question_df %>% mutate(patient_answer_text = patient_answer_text %>% str_replace_all(". NA", "."))
text_question_df = text_question_df %>% mutate(interviewer_question_text = interviewer_question_text %>% str_replace_all(". NA", "."))
text_question_df = text_question_df %>% select(-contains("volley")) %>% distinct(.keep_all = TRUE)

text_df = text_df %>% analyze_sentiment_from_sa_package(text=text_df$patient_text, colname_prefix = "patient_sentiment", include_only = "he")
text_df = text_df %>% analyze_sentiment_from_sa_package(text=text_df$interviewer_text, colname_prefix = "interviewer_sentiment", include_only = "he")

text_df$patient_sentiment_domainspecific = unlist(lapply(text_df$patient_text, score_domain_specific_sentiment, score_type="valence"))
text_df$interviewer_sentiment_domainspecific = unlist(lapply(text_df$interviewer_text, score_domain_specific_sentiment, score_type="valence"))

saveRDS(text_df, file="output/sentiment_at_role_level.RDS")
write.csv(text_df, file="output/sentiment_at_role_level.csv", row.names = FALSE)

text_volley_df = text_volley_df %>% analyze_sentiment_from_sa_package(text=text_volley_df$patient_volley_text, 
                                                                      colname_prefix = "patient_volley_sentiment",
                                                                      include_only = "he")
text_volley_df = text_volley_df %>% analyze_sentiment_from_sa_package(text=text_volley_df$interviewer_volley_text, 
                                                                      colname_prefix = "interviewer_volley_sentiment",
                                                                      include_only = "he")

text_volley_df$patient_volley_sentiment_domainspecific = unlist(lapply(text_volley_df$patient_volley_text, score_domain_specific_sentiment, score_type="valence"))
text_volley_df$interviewer_volley_sentiment_domainspecific = unlist(lapply(text_volley_df$interviewer_volley_text, score_domain_specific_sentiment, score_type="valence"))

text_volley_df = text_volley_df %>% mutate_all(function(x) ifelse(is.nan(x), NA, x))

saveRDS(text_volley_df, file="output/sentiment_at_volley_level.RDS")
write.csv(text_volley_df, file="output/sentiment_at_volley_level.csv", row.names = FALSE)

text_question_df = text_question_df %>% analyze_sentiment_from_sa_package(text=text_question_df$patient_answer_text, 
                                                                          colname_prefix = "patient_answer_sentiment",
                                                                          include_only = "he")
text_question_df = text_question_df %>% analyze_sentiment_from_sa_package(text=text_question_df$interviewer_question_text, 
                                                                          colname_prefix = "interviewer_question_sentiment",
                                                                          include_only = "he")



text_question_df$patient_answer_sentiment_domainspecific = unlist(lapply(text_question_df$patient_answer_text, score_domain_specific_sentiment, score_type="valence"))
text_question_df$interviewer_question_sentiment_domainspecific = unlist(lapply(text_question_df$interviewer_question_text, score_domain_specific_sentiment, score_type="valence"))

text_question_df = text_question_df %>% mutate_all(function(x) ifelse(is.nan(x), NA, x))

saveRDS(text_question_df, file="output/sentiment_at_question_level.RDS")
write.csv(text_question_df, file="output/sentiment_at_question_level.csv", row.names = FALSE)

