shhh = suppressMessages

shhh(library(tidyverse))
shhh(library(dplyr))
shhh(library(stringr))
shhh(library(tidyr))
shhh(library(topicmodels))
shhh(library(tidytext))
shhh(library(pdftools))
shhh(library(tm))
shhh(library(rvest))
shhh(library(SnowballC))
shhh(library(quanteda))
shhh(library(purrr))
shhh(library(ggwordcloud))
shhh(library(stm))
shhh(library(officer))
shhh(library(textreadr))
shhh(library(xml2))
shhh(library(SentimentAnalysis))
shhh(library(sentimentr))

# helper function for find.best.k.for.docs
normit = function(x){((x-min(x))/(max(x)-min(x)))+.000001}

mTurkWordBank = read_csv("dictionaries/wordbank_without_functions.csv")


valence_lexicon <- as.data.frame(lexicon::hash_valence_shifters)
valence_lexicon <- valence_lexicon %>% mutate(y = as.numeric(y))

polarity_lexicon = read_csv(file="data/polarity_lexicon.csv")
polarity_lexicon = polarity_lexicon %>% select(-lexicon)
colnames(polarity_lexicon) = c("x", "y")
valence_lexicon = valence_lexicon %>% anti_join(polarity_lexicon, by="x")


count_occurences_of_words = function(content_words, target_words)
{
  content_word_count = 0
  for (i in 1:length(content_words))
  {
    cur_word = content_words[i]
    for (j in 1:length(target_words))
    {
      cur_target_word = target_words[j]
      if (cur_word == cur_target_word)
      {
        content_word_count = content_word_count + 1
      }
    }
  }
  return(content_word_count)
}

score_sentiment_mturk <- function(content, min=10, max=0){
  content = tolower(content)
  content = gsub("[^[:alnum:] ]", "", content)
  scores = c()
  string_split <- strsplit(content, " ")[[1]]
  for (word in string_split)
  {
    word = as.character(word)
    if (length(which(mTurkWordBank$word==word)) > 0)
    {	
      index = which(mTurkWordBank$word==word)
      if ( (mTurkWordBank[index,]$happiness_average > max) |  (mTurkWordBank[index,]$happiness_average < min))
      {
        scores = c(scores, (mTurkWordBank[index,]$happiness_average))
      }
      else
      {
        scores = c(scores, NA)
      }
    }
    else
    {
      scores = c(scores, NA)
    }
  }
  return (mean(scores, na.rm=T))
}

score_polarity_mturk <- function(content, min=10, max=0){
  content = tolower(content)
  content = gsub("[^[:alnum:] ]", "", content)
  scores = c()
  string_split <- strsplit(content, " ")[[1]]
  for (word in string_split)
  {
    word = as.character(word)
    if (length(which(mTurkWordBank$word==word)) > 0)
    {	
      index = which(mTurkWordBank$word==word)
      if ( (mTurkWordBank[index,]$happiness_average > max) |  (mTurkWordBank[index,]$happiness_average < min))
      {
        scores = c(scores, (mTurkWordBank[index,]$happiness_average))
      }
      else
      {
        scores = c(scores, NA)
      }
    }
    else
    {
      scores = c(scores, NA)
    }
  }
  scores = abs(scores - 5)
  return (mean(scores, na.rm=T))
}

check_if_all_is_na = function(the_vector)
{
  for (i in length(the_vector))
  {
    cur_element = the_vector[i]
    if (cur_element %>% is.na() == FALSE)
    {
      return (FALSE)
    }
  }
  return (TRUE)
  
}

score_sentiment_sentimentr = function(content)
{
  content = tolower(content)
  sentiment_df = content %>% sentimentr::sentiment()
  sentiment_df = sentiment_df %>% mutate(weighted_sentiment = word_count * sentiment)
  overall_sentiment_df = sentiment_df %>% group_by(element_id) %>% summarise(weighted_avg_sentiment = sum(weighted_sentiment, na.rm=T)/sum(word_count))
  return(overall_sentiment_df$weighted_avg_sentiment[1])
}

score_domain_specific_sentiment = function(content, score_type="valence")
{
  content = tolower(content)
  
  if (score_type == "valence")
  {
    sentiment_df = content %>% sentimentr::sentiment_by(polarity_dt = polarity_lexicon %>% as_key(),
                                                        valence_shifters_dt = valence_lexicon %>% as_key(comparison = NULL))
    sentiment_score_for_content = sentiment_df$ave_sentiment[1]
    return(sentiment_score_for_content)
  }
  if (score_type == "no valence")
  {
    content_as_vector = strsplit(content, "\\s+")[[1]] %>% str_remove_all("[[:punct:]]")
    words_in_content = content_as_vector %>% length()
    positive_words = polarity_lexicon %>% filter(y == 1)
    positive_words = positive_words$x
    positive_word_count = positive_words %>% intersect(content_as_vector) %>% length()
    negative_words = polarity_lexicon %>% filter(y == -1)
    negative_words = negative_words$x
    negative_word_count = negative_words %>% intersect(content_as_vector) %>% length()
    sentiment_score_for_content = (positive_word_count - negative_word_count) / words_in_content
    return(sentiment_score_for_content)
  }
  if (score_type == "both")
  {

    sentiment_df = content %>% sentimentr::sentiment_by(polarity_dt = polarity_lexicon %>% as_key(),
                                                        valence_shifters_dt = valence_lexicon %>% as_key())
    sentiment_score_for_content = sentiment_df$ave_sentiment[1]
    if (sentiment_score_for_content == 0)
    {

      content_as_vector = strsplit(content, "\\s+")[[1]] %>% str_remove_all("[[:punct:]]")
      words_in_content = content_as_vector %>% length()
      positive_words = polarity_lexicon %>% filter(y == 1)
      positive_words = positive_words$x
      positive_word_count = count_occurences_of_words(content_as_vector, positive_words)
      negative_words = polarity_lexicon %>% filter(y == -1)
      negative_words = negative_words$x
      negative_word_count = count_occurences_of_words(content_as_vector, negative_words)
      sentiment_score_for_content = (positive_word_count - negative_word_count) / words_in_content
      # if (positive_word_count > 0 | negative_word_count > 0)
      # {
      #   print("-------------------------------------------")
      #   print(paste("content is:", content))
      #   print("valence sentiment_score_for_content is 0.")
      #   print(paste("positive words are:", positive_words %>% intersect(content_as_vector)))
      #   print(paste("negative words are:", negative_words %>% intersect(content_as_vector)))
      #   print(paste("positive_word_count is.", positive_word_count))
      #   print(paste("negative_word_count is.", negative_word_count))
      #   print(paste("non-valence sentiment_score_for_content is", sentiment_score_for_content))
      #   print("-------------------------------------------")
      # }
      
    }
    return(sentiment_score_for_content)
  }
}

score_sentiment_nrc = function(content)
{
  
  content = tolower(content)
  content = gsub("[^[:alnum:] ]", "", content)
  
  content_tibble <- strsplit(content, " ")[[1]] %>% tibble()
  colnames(content_tibble) = c("word")
  
  sentiment_tibble = get_sentiments("nrc")
  
  results_tibble = inner_join(content_tibble, sentiment_tibble, by=c("word"))
  
  #positive - negative word count divide by overall word count
  
  positive_word_count = results_tibble %>% filter(sentiment == "positive") %>% nrow()
  negative_word_count = results_tibble %>% filter(sentiment == "negative") %>% nrow()
  total_word_count = positive_word_count + negative_word_count
  
  if (total_word_count == 0)
  {
    return (NA)
  }
  else
  {
    total_sentiment = positive_word_count - negative_word_count
    avg_sentiment = total_sentiment / total_word_count
    return (avg_sentiment)
  }
  
}

score_sentiment_bing = function(content)
{
  
  content = tolower(content)
  content = gsub("[^[:alnum:] ]", "", content)
  
  content_tibble <- strsplit(content, " ")[[1]] %>% tibble()
  colnames(content_tibble) = c("word")
  
  sentiment_tibble = get_sentiments("bing")
  
  results_tibble = inner_join(content_tibble, sentiment_tibble, by=c("word"))
  
  #positive - negative word count divide by overall word count
  
  positive_word_count = results_tibble %>% filter(sentiment == "positive") %>% nrow()
  negative_word_count = results_tibble %>% filter(sentiment == "negative") %>% nrow()
  total_word_count = positive_word_count + negative_word_count
  
  if (total_word_count == 0)
  {
    return (NA)
  }
  else
  {
    total_sentiment = positive_word_count - negative_word_count
    avg_sentiment = total_sentiment / total_word_count
    return (avg_sentiment)
  }
  
}

score_sentiment_afinn = function(content)
{
  
  content = tolower(content)
  content = gsub("[^[:alnum:] ]", "", content)
  
  content_tibble <- strsplit(content, " ")[[1]] %>% tibble()
  colnames(content_tibble) = c("word")
  
  sentiment_tibble = get_sentiments("afinn")
  
  results_tibble = inner_join(content_tibble, sentiment_tibble, by=c("word"))
  avg_sentiment = mean(results_tibble$value, na.rm = TRUE)
  if (avg_sentiment %>% is.nan())
  {
    return (NA)
  }
  else
  {
    return (avg_sentiment)
  }
  
}

# helper function for find.best.k.for.docs
majority.vote = function(x) {
  ux <- unique(x)
  ux[which.max(tabulate(match(x, ux)))]
}

topic.model.from.pdf.to.tbl = function(pdf.file)
{
  success = FALSE
  try({
    pdf_txt = pdf_text(pdf.file)
    pdf_txt = pdf_txt %>% str_squish() %>% str_replace_all(",","")
    pdf_txt = unlist(strsplit(pdf_txt, "(?<=[[:punct:]])\\s(?=[A-Z])", perl=T))
    txt_this_iter = pdf_txt
    success = TRUE
  })
  if(success == TRUE)
  {
    cur_link = rep(pdf.file, length(pdf_txt))
    df = tibble(cur_link, pdf_txt)
    colnames(df) <- c('document', 'text')
    return(df)
  } else
  {
    df = data.frame(file=character(), text=character(), stringsAsFactors=FALSE) 
    return(df)
  }
  
}

topic.model.from.pptx.to.tbl = function(pptx.file)
{
  success = FALSE
  try({
    pptx_obj = read_pptx(pptx.file)
    pptx_tbl = pptx_summary(pptx_obj)
    pptx_tbl = pptx_tbl %>% filter(nchar(text) > 3)
    pptx_txt = pptx_tbl$text
    txt_this_iter = pptx_txt
    success = TRUE
  })
  if(success == TRUE)
  {
    cur_link = rep(pptx.file, length(pptx_txt))
    df = tibble(cur_link, pptx_txt)
    colnames(df) <- c('document', 'text')
    return(df)
  } else
  {
    df = data.frame(file=character(), text=character(), stringsAsFactors=FALSE) 
    return(df)
  }
  
}

topic.model.from.document.to.tbl = function(txt.file)
{
  success = FALSE
  try({
    the_txt = read_document(txt.file)
    the_txt = the_txt %>% str_squish() %>% str_replace_all(",","")
    the_txt = unlist(strsplit(the_txt, "(?<=[[:punct:]])\\s(?=[A-Z])", perl=T))
    txt_this_iter = the_txt
    success = TRUE
  })
  if(success == TRUE)
  {
    cur_link = rep(txt.file, length(the_txt))
    df = tibble(cur_link, the_txt)
    colnames(df) <- c('document', 'text')
    return(df)
  } else
  {
    df = data.frame(file=character(), text=character(), stringsAsFactors=FALSE) 
    return(df)
  }
  
}

topic.model.from.html.to.tbl = function(html.file)
{
  success = FALSE
  try({
    txt = xml2::read_html(html.file)
    p.txt = html_nodes(txt, "p")
    p.txt = html_text(p.txt) %>% str_squish() %>% str_replace_all(",","")
    p.txt = unlist(strsplit(p.txt, "(?<=[[:punct:]])\\s(?=[A-Z])", perl=T))
    
    h1.txt = html_nodes(txt, "h1")
    h1.txt = html_text(h1.txt) %>% str_squish() %>% str_replace_all(",","")
    h1.txt = unlist(strsplit(h1.txt, "(?<=[[:punct:]])\\s(?=[A-Z])", perl=T))
    
    h2.txt = html_nodes(txt, "h2")
    h2.txt = html_text(h2.txt) %>% str_squish() %>% str_replace_all(",","")
    h2.txt = unlist(strsplit(h2.txt, "(?<=[[:punct:]])\\s(?=[A-Z])", perl=T))
    
    h3.txt = html_nodes(txt, "h3")
    h3.txt = html_text(h3.txt) %>% str_squish() %>% str_replace_all(",","")
    h3.txt = unlist(strsplit(h3.txt, "(?<=[[:punct:]])\\s(?=[A-Z])", perl=T))
    
    h4.txt = html_nodes(txt, "h4")
    h4.txt = html_text(h4.txt) %>% str_squish() %>% str_replace_all(",","")
    h4.txt = unlist(strsplit(h4.txt, "(?<=[[:punct:]])\\s(?=[A-Z])", perl=T))
    
    h5.txt = html_nodes(txt, "h5")
    h5.txt = html_text(h5.txt) %>% str_squish() %>% str_replace_all(",","")
    h5.txt = unlist(strsplit(h5.txt, "(?<=[[:punct:]])\\s(?=[A-Z])", perl=T))
    
    h6.txt = html_nodes(txt, "h6")
    h6.txt = html_text(h6.txt) %>% str_squish() %>% str_replace_all(",","")
    h6.txt = unlist(strsplit(h6.txt, "(?<=[[:punct:]])\\s(?=[A-Z])", perl=T))
    
    txt_this_iter = c(p.txt, h1.txt, h2.txt, h3.txt, h4.txt, h5.txt, h6.txt)
    success = TRUE
  })
  
  if(success == TRUE)
  {
    cur_link <- rep(html.file, length(txt_this_iter))
    df = tibble(cur_link, txt_this_iter)
    colnames(df) <- c('document', 'text')
    return(df)
  }
  else
  {
    df = data.frame(file=character(), text=character(), stringsAsFactors=FALSE) 
    return(df)
  }
}

from.a.list.of.files.to.file.text.df = function(list.of.files)
{
  df = data.frame(file=character(), text=character(), stringsAsFactors=FALSE) 
  for (i in 1:length(list.of.files))
  {
    the.file = list.of.files[i]
    if (grepl(".pdf", the.file, ignore.case = TRUE) & file.exists(the.file)) {
      df.to.add = topic.model.from.pdf.to.tbl(the.file)
      df = df %>% rbind.data.frame(df.to.add)
    }
    else if (is.valid.textreadr.file.format(the.file) & file.exists(the.file)) {
      df.to.add = topic.model.from.document.to.tbl(the.file)
      df = df %>% rbind.data.frame(df.to.add)
    }
    else if (grepl(".pptx", the.file, ignore.case = TRUE) & file.exists(the.file)) {
      df.to.add = topic.model.from.pptx.to.tbl(the.file)
      df = df %>% rbind.data.frame(df.to.add)
    }
    else if (file.exists(the.file)) { # default right now largely for html
      df.to.add = topic.model.from.html.to.tbl(the.file)
      df = df %>% rbind.data.frame(df.to.add)
    }
  }
  # add a column based on file name that indicates id
  # df = df %>% mutate(ID = group_indices(., file))
  return(df)
}

get.tidy.text.from.dtm = function(the.doc.term.matrix)
{
  tidy.text.to.return = tidy(the.doc.term.matrix)
  colnames(tidy.text.to.return) = c("document", "word", "n")
  tidy.text.to.return = tidy.text.to.return %>% mutate(document=as.factor(document))
  return(tidy.text.to.return)
}

from.file.text.df.to.tidytext = function(file.text.df, remove.numbers=TRUE, remove.stop.words=TRUE, stem.words=TRUE)
{
  file.text.df.tokens = unnest_tokens(file.text.df, output = word, input = text)
  if (remove.numbers)
  {
    # remove numbers
    file.text.df.tokens = filter(file.text.df.tokens, !str_detect(word, "^[0-9]*$"))
  }
  if (remove.stop.words)
  {
    # remove stop words
    file.text.df.tokens = anti_join(file.text.df.tokens, stop_words)
  }
  if (stem.words)
  {
    # stem the words
    file.text.df.tokens = mutate(file.text.df.tokens, word = wordStem(word))
  }
  
  
  file.text.df.tokens = file.text.df.tokens %>% count(document, word, sort = TRUE) %>% ungroup()
  return(file.text.df.tokens)
}

from.tidy.text.to.dtm = function(tidy.text.tbl)
{
  dtm.to.return = tidy.text.tbl %>% cast_dtm(document, word, n)
  return(dtm.to.return)
}

get.lda = function(doc.term.matrix, k.parm, control.parm=1234)
{
  lda.to.return = LDA(doc.term.matrix, k = k.parm, control = list(seed = control.parm))
  return(lda.to.return)
}

get.tidy.topics.from.lda = function(lda.model, per.document=FALSE)
{
  if (per.document == FALSE)
  {
    topics.to.return =  tidy(lda.model, matrix = "beta")
  }
  else
  {
    topics.to.return =  tidy(lda.model, matrix = "gamma")
  }
  return(topics.to.return)
}


get.top.terms.from.topics = function(tidy.topics, top.n)
{
  top.terms.to.return = tidy.topics %>%  group_by(topic) %>% top_n(top.n, beta) %>% ungroup() %>% arrange(topic, -beta)
  return(top.terms.to.return)
}

get.word.cloud.from.tidy.text = function(tidy.text, per.document = FALSE, min.word.freq = 3, max.word.count = 100)
{
  tidy.text = tidy.text %>% filter(n >= min.word.freq)
  tidy.text = tidy.text %>% arrange(desc(n))
  
  if (per.document == FALSE)
  {
    tidy.text = tidy.text %>% mutate(count = sequence(n()))
    tidy.text = tidy.text %>% filter(count < max.word.count)
    # my.world.cloud = tidy.text %>% count(word) %>% with(wordcloud(word, n, min.freq =min.word.freq, max.words = max.word.count))
    my.world.cloud = ggplot(tidy.text, aes(label = word, size = n)) + geom_text_wordcloud_area() + scale_size_area(max_size = 20) + theme_minimal()
    return(my.world.cloud)
  } else
  {
    tidy.text = tidy.text %>% group_by(document) %>% mutate(count = sequence(n()))
    tidy.text = tidy.text %>% filter(count < max.word.count)
    my.world.cloud = ggplot(tidy.text, aes(label = word, size = n, color=document)) + geom_text_wordcloud_area() + scale_size_area(max_size = 20) + theme_minimal() + facet_wrap(~document)
    return(my.world.cloud)
  }
  
}

get.plot.for.top.terms = function(top.terms)
{
  plot.to.return = top.terms %>% mutate(term = reorder_within(term, beta, topic)) %>% 
    ggplot(aes(term, beta, fill = factor(topic))) +
    geom_col(show.legend = FALSE) +
    facet_wrap(~ topic, scales = "free") + xlab("word") + ylab("Importance to Topic") +
    coord_flip() +
    scale_x_reordered()
  
  return(plot.to.return)
}

get.plot.for.per.document = function(tidy.topics.per.doc)
{
  plot.to.return = tidy.topics.per.doc %>%
    mutate(document = reorder(document, gamma * topic)) %>%
    ggplot(aes(factor(topic), gamma)) +
    geom_boxplot() +
    facet_wrap(~ document)
  
  return(plot.to.return)
}

get.tidy.document.classification.from.lda = function(tidy.topics.per.doc)
{
  topic.classifications.to.return = tidy.topics.per.doc %>%
    group_by(document) %>%
    top_n(1, gamma) %>%
    ungroup()
  
  return(topic.classifications.to.return)
}

find.best.k.for.docs = function(docs.df, list.of.ks.to.try)
{
  tidy.docs.df <- docs.df %>%
    mutate(line = row_number()) %>%
    unnest_tokens(word, text) %>%
    anti_join(stop_words)
  
  tidy.docs.sparse <- tidy.docs.df %>%
    count(document, word, sort = TRUE) %>%
    cast_sparse(document, word, n)
  
  many_models <- tibble(K = list.of.ks.to.try) %>%
    mutate(topic_model = map(K, ~stm(tidy.docs.sparse, K = .,
                                     verbose = FALSE)))
  heldout <- make.heldout(tidy.docs.sparse)
  
  # if statement around this, if it errors out to few documents to return the min choice.
  successfully.gathered.metrics = FALSE
  try({
    k_result <- many_models %>%
      mutate(exclusivity = map(topic_model, exclusivity),
             semantic_coherence = map(topic_model, semanticCoherence, tidy.docs.sparse),
             eval_heldout = map(topic_model, eval.heldout, heldout$missing),
             residual = map(topic_model, checkResiduals, tidy.docs.sparse),
             bound =  map_dbl(topic_model, function(x) max(x$convergence$bound)),
             lfact = map_dbl(topic_model, function(x) lfactorial(x$settings$dim$K)),
             lbound = bound + lfact,
             iterations = map_dbl(topic_model, function(x) length(x$convergence$bound)))
    
    successfully.gathered.metrics = TRUE
    
  })
  
  if (successfully.gathered.metrics == FALSE)
  {
    return(list.of.ks.to.try %>% min())
  }
  
  k_result_tidy <- k_result %>%
    transmute(K,
              `Lower bound` = lbound,
              Residuals = map_dbl(residual, "dispersion"),
              `Semantic coherence` = map_dbl(semantic_coherence, mean),
              `Held-out likelihood` = map_dbl(eval_heldout, "expected.heldout")) %>%
    gather(Metric, Value, -K)
  
  k_result_tidy.hold.out = k_result_tidy %>% filter(Metric == "Held-out likelihood")
  index.best.hold.out = which.max(k_result_tidy.hold.out$Value)
  best.hold.out = k_result_tidy.hold.out$K[index.best.hold.out]
  
  k_result_tidy.residual = k_result_tidy %>% filter(Metric == "Residuals")
  index.best.residual= which.min(k_result_tidy.residual$Value)
  best.residual = k_result_tidy.residual$K[index.best.residual]
  
  k_exclusivity <- k_result %>%
    select(K, exclusivity, semantic_coherence) %>%
    filter(K %in% list.of.ks.to.try) %>%
    unnest() %>%
    mutate(norm.exclusivity = normit(exclusivity)) %>% 
    mutate(norm.semantic_coherence = normit(semantic_coherence)) %>% 
    mutate(norm.score = sqrt(norm.exclusivity*norm.semantic_coherence))
  
  # we want ks with high hold-out likelihood, low residuals, and high semantic coherence and high exclusivity
  index.best.exclusivity.coherence.balance = which.max(k_exclusivity$norm.score)
  best.exclusivity.coherence.balance = k_exclusivity$K[index.best.exclusivity.coherence.balance]
  
  # take the most common choice, if there is a common choice take the min
  list.of.choices = c(best.hold.out, best.residual, best.exclusivity.coherence.balance) %>% sort()
  the.choice.for.k = majority.vote(list.of.choices)
  return(the.choice.for.k)
}

get.top.tfidf.terms.from.document = function(tidy.text.words.tfidf, top.n)
{
  top.tidy.text.words.tfidf = tidy.text.words.tfidf %>% top_n(top.n, tf_idf) %>% arrange(-tf_idf)
  # this would force only top.n numbers even in case of ties, for now we allow ties
  # top.tidy.text.words.tfidf = top.tidy.text.words.tfidf[1:top.n,]
  return(top.tidy.text.words.tfidf)
}

get.top.tfidf.terms.plot.from.document = function(tidy.text.words.tfidf.top.n)
{
  
  my.plot = tidy.text.words.tfidf.top.n  %>% ggplot(aes(x=reorder(word, tf_idf), y=tf_idf, fill = factor(word))) +
    geom_col(show.legend = FALSE) + coord_flip() + scale_x_reordered() + xlab("word") + ylab("Uniqueness of words in doc relative to other web results.")
  return(my.plot)
}

get.top.terms.from.document = function(tidy.text.df, top.n)
{
  top.terms.to.return = tidy.text.df %>% top_n(top.n, n) %>% arrange(-n)
  return(top.terms.to.return)
}

get.top.n.terms.plot.from.tidy.text = function(tidy.text.df.of.top.n)
{
  
  my.plot = tidy.text.df.of.top.n  %>% ggplot(aes(x=reorder(word, n), y=n, fill = factor(word))) +
    geom_col(show.legend = FALSE) + coord_flip() + scale_x_reordered() + xlab("word") + ylab("Frequency")
  return(my.plot)
}





